import logging
import mimetypes
from datetime import timedelta
from io import BytesIO
from pathlib import Path

import gcsfs
import requests
from ...cloud_path import CloudPath
from ...cloud_path import CloudPathFactory
from ...decor import gcp_exception
from tqdm import tqdm  # type: ignore

try:
    from google.cloud import storage  # type: ignore
    from google.cloud.storage import Bucket  # type: ignore
    import google.cloud.storage.constants as StorageApiConstants  # type: ignore
except ImportError:
    raise ImportError(
        "Please install google-cloud-storage using pip install google-cloud-storage"
    )

log = logging.getLogger("rich")

@CloudPathFactory.register("gs")
class GCPCloudPath(CloudPath):
    """
    This class creates a GCP CloudFile object from a path.
    """

    def __init__(self, cloud_path: str, *args, **kwargs) -> None:
        """
        Constructor
        :param cloud_path: Cloud path
        """
        super().__init__(cloud_path)
        self._client = kwargs.get("client", storage.Client())

    def get_mapper(self, *args, **kwargs):
        """
        Get the mapper
        """
        gcs = gcsfs.GCSFileSystem()
        store = gcs.get_mapper(self.as_uri(), *args, **kwargs)
        return store

    @property
    def bucket_name(self) -> str:
        """
        Get the bucket name
        """
        return self.uri_parts.netloc

    def as_uri(self):
        """
        Get the URI of the path
        """
        return self.uri

    @property
    def blob_name(self) -> str:
        """
        Get the blob name
        """
        return self.uri_parts.path.lstrip("/")

    @property
    def bucket(self) -> Bucket:
        """
        Get the bucket
        """
        return self._client.bucket(self.bucket_name)

    @property
    def blob(self):
        """
        Get the blob  object
        """
        bucket = self.bucket
        blob_name = self.blob_name
        return bucket.blob(blob_name)

    @property
    def metadata(self):
        """
        Get metadata
        :return: Metadata
        """
        return self.blob.metadata

    @property
    def content_type(self) -> str:
        """
        Get file content type.
        :return: Content type
        """
        bucket = self._client.get_bucket(self.bucket_name)
        blob = bucket.get_blob(self.blob_name)
        return blob.content_type

    @metadata.setter
    def metadata(self, metadata: dict):
        """
        Set metadata
        :param metadata: Metadata
        """
        blob = self.blob
        blob.metadata = metadata
        blob.patch()

    @gcp_exception
    def get_tags(self):
        """
        Get tags
        :return: Tags
        """
        blob = self.blob
        return {
            "etag": blob.etag,
            "size": blob.size,
            "updated": blob.updated,
            "content_type": blob.content_type,
        }

    def _create_bucket(
        self,
        bucket_class=StorageApiConstants.STANDARD_STORAGE_CLASS,
        location="us",
    ):
        """
        this function creates a new bucket in the google cloud storage
        :param bucket_class:
        :return:
        """
        bucket = self.bucket
        if bucket.exists():
            return bucket
        bucket.storage_class = bucket_class
        bucket.location = location
        new_bucket = bucket.create()
        return new_bucket

    def _delete_bucket(self):
        """
        this function deletes a bucket from the google cloud storage
        :return:
        """
        bucket = self.bucket
        if not bucket.exists():
            return
        # List all objects in the bucket and delete each one
        blobs = bucket.list_blobs()
        for blob in blobs:
            blob.delete()
        bucket.delete()

    @gcp_exception
    def exists(self) -> bool:
        """
        Check if the specified path exists in the GCP bucket.
        For a folder path, verifies if there are any objects with the given prefix.
        For a file path, checks if the specific blob exists.

        :return: True if the path exists (file or folder), False otherwise.
        """
        if self.is_folder_path():
            # List only the first object in the "folder" to improve performance
            return any(self.bucket.list_blobs(prefix=self.blob_name, max_results=1))
        else:
            # Directly check if the blob (file) exists
            return self.blob.exists()

    def is_folder_path(self) -> bool:
        """
        Determine if the specified path is intended to represent a folder.

        :return: True if the path ends with a "/", indicating a folder, False otherwise.
        """
        return self.uri_parts.path.endswith("/")

    @gcp_exception
    def is_dir(self) -> bool:
        """
        Check if the specified path is an existing directory in the GCP bucket.

        :return: True if the path is a folder and contains at least one object,
                 False otherwise.
        """
        return self.is_folder_path() and self.exists()

    @gcp_exception
    def is_file(self):
        """
        Check if the path is a file
        :return: True if the path is a file, False otherwise
        """
        blob_path = self.uri_parts.path.lstrip("/")
        return not blob_path.endswith("/") and self.blob.exists()

    @gcp_exception
    def mkdir(self, exist_ok: bool = False, parents: bool = False) -> None:
        """
        Create a directory
        :param exist_ok: If True, do not raise an exception if the directory
        already exists
        :param parents: If True, also create any missing parent directories
        :return: None
        """
        bucket = self.bucket
        blob_path = self.uri_parts.path.lstrip("/")
        if not bucket.exists():
            self._create_bucket()
        if blob_path == "":
            return

        assert blob_path.endswith(
            "/"
        ), "you are attempting to create a directory, make sure the path ends with a /"

        blob = self.blob
        # check if the folder has already been created
        if blob.exists():
            if not exist_ok:
                raise FileExistsError(f"Directory exists: {self}")
            return

        blob.upload_from_string(
            "", content_type="application/x-www-form-urlencoded;charset=UTF-8"
        )

    @gcp_exception
    def rmdir(self):
        """
        Remove the directory
        :return: None
        """
        blob_path = self.uri_parts.path.lstrip("/")
        if not self.is_dir():
            raise NotADirectoryError(f"Not a directory: {self}")

        blobs = self.bucket.list_blobs(prefix=blob_path)
        for blob in blobs:
            blob.delete()

    @gcp_exception
    def read_bytes(self) -> bytes:
        """
        Read the file as bytes
        :return: File as bytes
        """
        if self.is_dir():
            raise IsADirectoryError(f"Is a directory: {self}")

        signed_url = self.blob.generate_signed_url(
            version="v4", expiration=timedelta(minutes=15), method="GET"
        )
        response = requests.get(signed_url, stream=True)
        file_size = int(response.headers.get("content-length", 0))
        file_chunk_size = 1024 * 1024 * 10  # 10 MB
        progress_bar = tqdm(
            total=file_size, unit="iB", unit_scale=True, desc="Downloading file"
        )
        with BytesIO() as data_stream:
            for data in response.iter_content(file_chunk_size):
                progress_bar.update(len(data))
                data_stream.write(data)
            progress_bar.close()
            if file_size != 0 and progress_bar.n != file_size:
                raise ValueError("Error in downloading file from GCP")
            return data_stream.getvalue()

    @gcp_exception
    def write_bytes(self, data: bytes):
        """
        Write bytes to a file
        :param data: Data to write
        already exists
        :return: None
        """
        if self.is_dir():
            raise IsADirectoryError(f"Is a directory: {self}")
        blob_content_type = mimetypes.guess_type(str(self))[0]
        signed_url = self.blob.create_resumable_upload_session(
            content_type=blob_content_type, client=self._client
        )
        data_length = len(data)
        file_chunk_size = 1024 * 1024 * 10

        with tqdm(
            total=data_length, unit="B", unit_scale=True, desc="Uploading file"
        ) as progress_bar:
            for i in range(0, data_length, file_chunk_size):
                chunk = data[i : i + file_chunk_size]
                start_byte_index = i
                end_byte_index = i + len(chunk)
                response = requests.put(
                    signed_url,
                    data=chunk,
                    headers={
                        "Content-Type": "application/octet-stream",
                        "Content-Range": f"bytes {start_byte_index}-{end_byte_index - 1}/{data_length}",
                    },
                )
                if response.status_code == 308:
                    progress_bar.update(len(chunk))
                elif response.status_code == 200:
                    # log.info("File uploaded successfully")
                    break

    @gcp_exception
    def iterdir(self):
        """
        Iterate over the directory
        :return: Iterator
        """
        blob_path = self.uri_parts.path.lstrip("/")
        if blob_path == "":
            for blob in self.bucket.list_blobs():
                yield GCPCloudPath(f"gs://{self.uri_parts.netloc}/{blob.name}")

        if not self.is_dir():
            raise NotADirectoryError(f"Not a directory: {self}")

        blob_path = self.uri_parts.path.lstrip("/")
        bucket_name = self.uri_parts.netloc
        for child_blob in self.bucket.list_blobs(prefix=blob_path):
            yield GCPCloudPath(f"gs://{bucket_name}/{child_blob.name}")

    def __fspath__(self):
        """
        Get the file system path
        :return: File system path
        """

        if self.is_dir():
            raise IsADirectoryError(f"Is a directory: {self}")
        self.__check_for_file_in_cache()
        return super().__fspath__()

    def __download_file_to_cache(self):
        """
        Download the file to the cache
        :return:
        """
        # transform the cloud path to a local path
        cloud_file_local_path = Path(self)
        # create the parent directory where the file will be downloaded if it does not exist
        cloud_file_local_path.parent.mkdir(parents=True, exist_ok=True)
        # download the file from the cloud and load it into memory
        cloud_file_bytes = self.read_bytes()
        # write the file to the local path
        with open(cloud_file_local_path, "wb") as file_obj:
            file_obj.write(cloud_file_bytes)

    def __check_for_file_in_cache(self):
        """
        Get or update the cached file, the file is updated if it is older than 1 day
        :return:
        """
        cloud_file_local_path = Path(self)
        cloud_file_is_cached = cloud_file_local_path.exists()
        if cloud_file_is_cached:
            cloud_file_last_modified = cloud_file_local_path.stat().st_mtime
            cloud_file_age = timedelta(seconds=cloud_file_last_modified)
            cloud_file_age_threshold = timedelta(days=1)
            cloud_file_is_outdated = cloud_file_age < cloud_file_age_threshold
            if cloud_file_is_outdated:
                self.__download_file_to_cache()
        else:
            self.__download_file_to_cache()



class CloudFsPyException(Exception):
    """
    Base class for CloudFs
    """


class NotSchemeSupported(CloudFsPyException):
    """
    Exception raised when a cloud provider is not supported.
    """


class InvalidCloudPath(CloudFsPyException):
    """
    Exception raised when a cloud path is invalid.
    """



import os
import pathlib
import re
import typing
from urllib.parse import urlparse


def get_uri_scheme(url) -> typing.Optional[str]:
    """
    Get the scheme from a URL.
    :param url:  URL
    :return: Scheme
    """
    pattern = r"^(.+?)://"
    match = re.match(pattern, url)
    if match:
        return match.group(1)  # Return the matched scheme
    return None


def is_valid_uri(uri: str) -> bool:
    """
    Check if a URI is valid.
    :param uri: URI
    :return: True if the URI is valid, False otherwise
    """
    try:
        result = urlparse(uri)
        # Check if the scheme and netloc are present
        return all([result.scheme, result.netloc])
    except:  # noqa
        return False


def rm_tree(pth):
    """
    Remove a directory and its contents.
    :param pth:
    :return:
    """
    pth = pathlib.Path(pth)
    for child in pth.glob("*"):
        if child.is_file():
            child.unlink()
        else:
            rm_tree(child)
    pth.rmdir()


def ls_tree(startpath, max_levels, symbol="-", indent=0):
    """
    List the contents of a directory tree.
    :param startpath:
    :param max_levels:
    :param symbol:
    :param indent:
    :return:
    """
    if max_levels < 0:  # Base case to stop recursion
        return ""

    output = ""
    # Use the provided symbol as the indicator for each level
    prefix = "    " * indent + (symbol + " " if indent > 0 else "")
    output += prefix + os.path.basename(startpath) + "\n"
    prefix = "    " * (indent + 1) + symbol + " "

    if indent >= max_levels:  # Stop diving into directories if max depth is reached
        return output

    try:
        for item in os.listdir(startpath):
            path = os.path.join(startpath, item)
            if os.path.isdir(path):
                output += ls_tree(path, max_levels - 1, symbol, indent + 1)
            else:
                output += prefix + item + "\n"
    except PermissionError:
        output += prefix + "Permission denied" + "\n"

    return output



import logging
import sys
import traceback
from functools import wraps

log = logging.getLogger("rich")


def exception(function):
    """
    Decorator to log exceptions.
    :param function: function to decorate.
    :return: decorated function.
    """

    @wraps(function)
    def wrapper(*args, **kwargs):
        """
        Wrapper function to log exceptions.
        :param args:
        :param kwargs:
        :return:
        """
        try:
            return function(*args, **kwargs)
        except Exception:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            exc_message = traceback.format_exception_only(exc_type, exc_value)[0]
            log.exception(
                " Calling the function `{}` -> {} ".format(
                    function.__qualname__, exc_message
                )
            )
            raise

    return wrapper


def gcp_exception(function):
    """
    Decorator to log exceptions.
    :param function: function to decorate.
    :return:
    """
    try:
        from google.cloud.exceptions import NotFound, BadRequest, Forbidden  # type: ignore
    except ImportError:
        raise ImportError(
            "Please install google-cloud-storage using pip install google-cloud-storage"
        )

    @wraps(function)
    def wrapper(self, *args, **kwargs):
        """
        Wrapper function to catch exceptions.
        :param self:
        :param args:
        :param kwargs:
        :return:
        """
        try:
            return function(self, *args, **kwargs)
        except (NotFound, BadRequest, Forbidden, Exception) as ex:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            exc_message = traceback.format_exception_only(exc_type, exc_value)[0]

            error_messages = {
                NotFound: "bucket not found",
                BadRequest: "Invalid request to the gcp file",
                Forbidden: "Access to the gcp path is forbidden",
            }
            log.exception(
                f"Calling the function `{function.__qualname__}` "
                f"-> {error_messages.get(type(ex), 'Unknown error')} : {exc_message}"
            )
            raise

    return wrapper



import os
import typing
from pathlib import _posix_flavour  # type: ignore
from pathlib import _windows_flavour  # type: ignore
from pathlib import Path as _Path_

if typing.TYPE_CHECKING:
    from . import GenericPath
    from .local_path import LocalPath
    from .cloud_path import CloudPath


class AnyPath(_Path_):
    """
    GenericPath class is an abstract class
    """

    _flavour = _windows_flavour if os.name == "nt" else _posix_flavour

    def upload_to(
        self,
        destination: typing.Union[str, "GenericPath", "CloudPath"],
        *args,
        **kwargs,
    ) -> "CloudPath":
        """
        Upload a file to a destination.
        :param destination: Destination path
        """
        raise NotImplementedError

    def get_mapper(self, *args, **kwargs):
        """
        Get the mapper
        :param args: Arguments
        :param kwargs: Keyword arguments
        :return: Mapper
        """
        raise NotImplementedError

    def download_to(
        self,
        destination: typing.Union[str, "GenericPath", "LocalPath"],
        *args,
        **kwargs,
    ) -> "LocalPath":
        """
        Download a file to a destination.
        :param destination: Destination path
        """
        raise NotImplementedError



import os
import tempfile
from pathlib import Path
import skyio


from .utils import rm_tree, ls_tree

CACHE_FOLDER_ENV_VAR_NAME = skyio.__name__.upper() + "_CACHE"


def set_cache_folder(folder: str):
    """
    Set the cache folder.
    :param folder: The folder to set.
    """
    os.environ[CACHE_FOLDER_ENV_VAR_NAME] = folder


def get_cache_folder() -> Path:
    """
    Get the cache folder.
    :return: The cache folder.
    """
    default_location = Path(tempfile.gettempdir()).joinpath(f".{skyio.__name__}")
    folder = Path(os.environ.get(CACHE_FOLDER_ENV_VAR_NAME, default_location))
    set_cache_folder(str(folder))
    folder.mkdir(exist_ok=True, parents=True)
    return folder


def clear_cache_folder():
    """
    Clear the cache.
    """
    cache_folder = get_cache_folder()
    rm_tree(cache_folder)


def list_cache_folder(max_levels=2):
    """
    List the cache.
    """
    cache_folder = get_cache_folder()
    print(ls_tree(cache_folder, max_levels=max_levels))



#!/usr/bin/env python
# coding: utf-8

# # Getting started with cloudfspy
# 
# `cloudFspy` provides an api based on `pathlib.Path` to access files on cloud storage services. It is a wrapper, similar to `fsspec`, around the `fsspec` library, which provides a common interface to many different file systems. `cloudFspy` is designed to be easy to use and to provide a consistent interface to access files on cloud storage services.This tutorial will show you how to use `cloudFspy` to access files on cloud storage services.
# 
# ### Downloading and Uploading Files
# 
# We start by importing the `GenericPath` class from the `cloudFspy` library. We can then use the `GenericPath` class to access files on cloud storage services.

# In[1]:


import cloudfspy
from cloudfspy import GenericPath as Path


# The `set_cache_folder` and `clear_cache_folder` methods enable the configuration and purging of the cache directory. This directory serves as a temporary storage area for files fetched from the cloud, facilitating efficient access to large files without the need for repeated downloads. Future enhancements will introduce various caching tactics. Presently, our caching approach involves transferring the file to the cache directory for subsequent access. If a file in the cache is more than two days old, it will be re-downloaded. We aim to make this behavior adjustable in upcoming updates.

# In[2]:


# set the cache folder
cloudfspy.set_cache_folder("cache")
cloudfspy.clear_cache_folder()


# **Downloading files**
# 

# In[3]:


from_path = Path("gs://gpr-studio-data/Archive.zip")
to_path = Path("/Users/haruiz/Desktop/Archive.zip")

from_path.download_to(to_path)
with open(to_path, "rb") as f:
    data = f.read()


# **Uploading Files**

# In[4]:


from_path = Path("/Users/haruiz/Desktop/Archive.zip")
to_path = Path("gs://gpr-studio-data/Archive.zip")

from_path.upload_to(to_path)
with open(to_path, "rb") as f:
    data = f.read()


# In[5]:


cloudfspy.list_cache_folder(max_levels=5)




import logging

from rich.logging import RichHandler  # type: ignore

from .cloud_path import CloudPath
from .cloud_path import CloudPathFactory
from .generic_path import GenericPath
from .local_path import LocalPath
from .settings import clear_cache_folder
from .settings import get_cache_folder
from .settings import set_cache_folder
from .vendors import GCPCloudPath

FORMAT = "%(message)s"
logging.basicConfig(
    level=logging.INFO,
    format=FORMAT,
    datefmt="[%X]",
    handlers=[RichHandler(rich_tracebacks=False)],
)

__version__ = "0.1.0"
__name__ = "skyio"



import inspect
import logging
import typing
from functools import wraps
from inspect import signature
from typing import get_args
from typing import get_origin

log = logging.getLogger("rich")


def typeassert(func):
    """
    Decorator to check the type of the arguments of a function.
    :param func: function to decorate.
    :return:
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        func_sig = signature(func)
        func_args_signature = func_sig.parameters
        func_args = func_sig.bind(*args, **kwargs).arguments
        func_args.pop("self", None)
        for arg_name, arg_value in func_args.items():
            arg_signature = func_args_signature[arg_name]
            arg_annotation = arg_signature.annotation

            if arg_annotation is inspect.Parameter.empty or arg_value is None:
                continue

            if (
                isinstance(arg_annotation, list)
                or get_origin(arg_annotation) == typing.Union
            ):
                valid_types = (
                    arg_annotation
                    if isinstance(arg_annotation, list)
                    else list(get_args(arg_annotation))
                )

                type_checking = [isinstance(arg_value, t) for t in valid_types]
                if not any(type_checking):
                    supported_types = ",".join(list(map(str, valid_types)))
                    log.error(
                        f"The argument `{arg_name}` of the function `{func.__qualname__}` must be one of the "
                        f"following types: {supported_types}"
                    )
                    raise TypeError

            else:
                if not isinstance(arg_value, arg_annotation):
                    log.error(
                        f"argument `{arg_name}` must be `{arg_annotation}` in the function `{func.__qualname__}`"
                    )
                    raise TypeError
        return func(*args, **kwargs)

    return wrapper



import typing
from pathlib import Path
from typing import Callable
from typing import Dict
from urllib.parse import urlparse

from .any_path import AnyPath
from .exceptions import NotSchemeSupported
from .settings import get_cache_folder

if typing.TYPE_CHECKING:
    from .local_path import LocalPath
    from .generic_path import GenericPath


class CloudPathFactory:
    """
    This class creates a CloudFile object from a path.
    """

    registry: Dict[str, AnyPath] = {}

    @classmethod
    def register(cls, key: str) -> Callable:
        """
        Register a file handler class.
        :param key: Key for the file handler class
        :return: File handler class
        """

        def wrapper(wrapped_class: AnyPath) -> AnyPath:
            """
            Register a file handler class
            :param wrapped_class:  File handler class
            :return:
            """
            cls.registry[key] = wrapped_class
            return wrapped_class

        return wrapper

    @classmethod
    def build(cls, key: str) -> AnyPath:
        """
        Build a file handler instance for the given key.
        :param key: Key for the file handler class
        :param kwargs: The keyword arguments to pass to the file
        handler constructor.
        :return: File handler instance
        """
        if key not in cls.registry:
            raise NotSchemeSupported(f"scheme {key} not supported")
        file_handler_class = cls.registry[key]
        return file_handler_class


class CloudPath(AnyPath):
    """
    CloudPath class is a subclass of GenericPath
    """

    def __new__(cls, cloud_path: str, *args, **kwargs):
        """
        Create a CloudPath object from a path.
        :param args: Arguments
        :param kwargs: Keyword arguments
        """
        cloud_path_parts = urlparse(cloud_path)
        path_scheme = cloud_path_parts.scheme
        path_netloc = cloud_path_parts.netloc
        cloud_path_cls = CloudPathFactory.build(path_scheme)
        # the local path is the cache folder + the netloc + the path

        cache_folder = get_cache_folder()
        local_path = cache_folder.joinpath(
            path_netloc, cloud_path_parts.path.lstrip("/")
        )
        return super().__new__(cloud_path_cls, local_path, *args, **kwargs)  # type: ignore

    def __init__(self, cloud_path: str, *args, **kwargs):
        """
        Constructor
        :param args: Arguments
        :param kwargs: Keyword arguments
        """
        super().__init__()

        self.uri = cloud_path
        self.uri_parts = urlparse(cloud_path)

    @property
    def content_type(self):
        """
        Get content type
        :return: Content type
        """
        raise NotImplementedError

    @property
    def scheme(self):
        """
        Get the scheme
        :return: Scheme
        """
        return self.uri_parts.scheme

    @property
    def parent(self):
        """
        Get the parent path
        :return: Parent path
        """
        parent_path = Path(self.uri_parts.path).parent
        parent_uri = self.drive + self.uri_parts.netloc + str(parent_path)
        cls = self.__class__
        if parent_path == Path("/"):
            return cls(parent_uri)
        return cls(parent_uri)

    @property
    def metadata(self):
        """
        Get metadata
        :return: Metadata
        """
        raise NotImplementedError

    @metadata.setter
    def metadata(self, metadata: dict):
        """
        Set metadata
        :param metadata: Metadata
        """
        raise NotImplementedError

    def joinpath(self, *paths):
        """
        Join the path
        :param other: Other path
        :return: New path
        """
        cache_folder = get_cache_folder()
        path_objects = [self] + [Path(p) for p in paths]
        join_path = Path.joinpath(*path_objects).relative_to(cache_folder)
        new_uri = self.drive + str(join_path)
        if paths[-1].endswith("/"):
            new_uri += "/"
        return self.__class__(new_uri)

    @property
    def drive(self):
        """
        Get the GCP drive
        :return: GCP drive
        """
        return self.uri_parts.scheme + "://"

    @property
    def root(self):
        """
        Get the root
        :return: Root
        """
        return self.drive + self.uri_parts.netloc + "/"

    def download_to(
        self,
        destination: typing.Union[str, "GenericPath", "AnyPath", "LocalPath"],
        *args,
        **kwargs,
    ) -> "LocalPath":
        """
        Download a file to a destination.
        :param destination: Destination path (LocalPath)
        """
        from .local_path import LocalPath

        assert isinstance(
            destination, (str, LocalPath)
        ), "Invalid destination parameter"
        if isinstance(destination, str):
            destination = LocalPath(destination)
        with open(destination, "wb+") as f:
            # read bytes from cloud and write to local
            f.write(self.read_bytes())

        return destination



import typing
from pathlib import Path

from .any_path import AnyPath
from .cloud_path import CloudPath
from .local_path import LocalPath
from .utils import is_valid_uri


class GenericPath(AnyPath):
    """
    Generic path class.
    """

    def __new__(cls, path: typing.Union[str, Path], *args, **kwargs):
        """
        Create a CloudPath object from a path.
        :param path: Path
        :param args: Arguments
        :param kwargs: Keyword arguments
        """
        # copy docs from Path class
        if isinstance(path, AnyPath):
            return path

        path = str(path)
        if is_valid_uri(path):
            return CloudPath(path, *args, **kwargs)
        else:
            return LocalPath(path, *args, **kwargs)



import typing

from .any_path import AnyPath

if typing.TYPE_CHECKING:
    from .cloud_path import CloudPath
    from .generic_path import GenericPath


class LocalPath(AnyPath):
    def __init__(self, *args, **kwargs):
        """
        Constructor
        :param args: Arguments
        :param kwargs: Keyword arguments
        """
        super().__init__()

    def get_mapper(self, *args, **kwargs):
        """
        Get the mapper
        :param args: Arguments
        :param kwargs: Keyword arguments
        :return: Mapper
        """
        return self

    def upload_to(
        self,
        destination: typing.Union[str, "GenericPath", "CloudPath"],
        *args,
        **kwargs,
    ) -> "CloudPath":
        """
        Upload a file to a destination.
        :param destination: Destination path (CloudPath)
        """
        from .cloud_path import CloudPath

        assert isinstance(
            destination, (str, CloudPath)
        ), "Invalid destination parameter"
        if isinstance(destination, str):
            destination = CloudPath(destination)

        with open(self, "rb") as f:
            # read bytes from local and write to cloud
            destination.write_bytes(f.read())

        metadata = kwargs.get("metadata")
        if metadata is not None and isinstance(metadata, dict):
            destination.metadata = metadata

        return destination
